#include <iostream>
using std::cout;
#include <iomanip>
using std::endl;
#include <cmath>
using std::pow;
using std::sqrt;
using std::exp;
#include "dvegas.h"
using HepSource::STRATIFIED;

extern "C" { void trapfpe(); }  // guarantee abort on floating point exception

// define test integrands ---------------------------------------

// no adaptation ************************************************

// helper:
const double normalizedPower(const double& x, int exponent = 4)
{
  return pow(x, exponent) * (exponent+1);
}

void crudeMC(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  f[0] = normalizedPower(aux[0]);
}

void crudeMCMultipleIntegrands(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  for (int i = 1; i <= 5; ++i)
    f[i-1] = pow(aux[0], i);
}

// with adaptation **********************************************

void adaptiveMC(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  f[0] = normalizedPower(x[0]);
}

void adaptiveMCMultipleIntegrands(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  for (int i = 1; i <= 5; ++i)
    f[i-1] = pow(x[0], i);
  // adaptation only takes first integrand, i.e. f[0], into account
}

// Lepage: J. Comp. Phys. 27, 192 ------

// helper:
const double sphericallySymmetricGaussian(const double x[], const double& offset)
{
  const double a = 0.1;
  const int n = 4;     // dimension
  const double sqrtPi = 1.772453850905516;
  double sum = 0.0;
  for (int i = 0; i < n; ++i)
    sum += pow((x[i] - offset)/a, 2);
  return 1./pow(a*sqrtPi, n) * exp(-sum);
}

void lepageEq8(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  f[0] = sphericallySymmetricGaussian(x, 0.5);
}

void crudeLepageEq8(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  f[0] = sphericallySymmetricGaussian(aux, 0.5);
}

void lepageEq9(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  f[0] = 0.5 * (sphericallySymmetricGaussian(x, 1./3)
                + sphericallySymmetricGaussian(x, 2./3));
}

// -------------------------------------

// optimizing sums of integrals *********************************

void adaptationForSumsAndAverages(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  f[0] = ((k[0] == 0) ? 100. : 1.)/(100.+99.*1.) * normalizedPower(x[0]);
}

void crudeForSumsAndAverages(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  f[0] = ((aux[0] < 0.01) ? 100. : 1.)*100./(100.+99.*1.) * normalizedPower(x[0]);
}

// adaptation that takes correlations into account **************

// helper:
const double nastyForVEGAS(const double x[])
{
  // volume of diagonal cylinder in unit cube
  const double a = 0.05;
  const double radius = 0.5 * sqrt(2./3) * a;
  const double lengthDiagonal = sqrt(3.);
  const double heightPyramid = a/sqrt(3.);
  const double height = lengthDiagonal - 2. * heightPyramid;
  const double volume = 3.14159265358979 * pow(radius, 2) * height;

  const double distOrigin = sqrt(pow(x[0], 2) + pow(x[1], 2) + pow(x[2], 2));
  const double cosThetaDiagonal = (x[0] * 1. + x[1] * 1. + x[2] * 1.)
    / lengthDiagonal / distOrigin;
  const double sinThetaDiagonal = sqrt(1. - pow(cosThetaDiagonal, 2));
  const double heightX = distOrigin * cosThetaDiagonal - heightPyramid;
  const double radiusX = distOrigin * sinThetaDiagonal;

  const double normFac = 1./(1. * volume + 0.001 * (1. - volume));
  return normFac*((0. < heightX && heightX < height && radiusX < radius) ? 1. : 0.001);
}

void nastyCrudeMC(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  f[0] = nastyForVEGAS(aux);
}

// take into account correlations of 0, 1, 2 and 3 dimensions
void nastyWithAdaptation(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  f[0] = nastyForVEGAS(x);
}

void twoSetsOfCorrelatedDimensions(const double x[], const int k[], const double& wgt, const double aux[], double f[])
{
  f[0] = nastyForVEGAS(x) * nastyForVEGAS(&x[3]);     // x[6]
}

// --------------------------------------------------------------

int main(int argc, char* argv[])
{
  // correct results are 1.0 unless specified otherwise

  using HepSource::Dvegas;
  using HepSource::VEGAS;
  trapfpe();

  cout << endl << "+++++ no adaptation" << endl << endl;

  Dvegas dv1(1, 1, &crudeMC);
  VEGAS(dv1, 10000, 2);

  cout << endl;
  
  Dvegas dv2(1, 5, &crudeMCMultipleIntegrands);
  VEGAS(dv2, 10000, 2);     // results: 1/(i+1) for i = 1, 2, ..., 5

  cout << endl << "+++++ with adaptation" << endl << endl;

  Dvegas dv3(1, 50, 1, &adaptiveMC);
  VEGAS(dv3, 10000, 3);
  
  cout << endl;
  
  Dvegas dv4(1, 50, 5, &adaptiveMCMultipleIntegrands);
  VEGAS(dv4, 10000, 3);     // results: 1/(i+1) for i = 1, 2, ..., 5

  cout << endl << "+++++ Gaussian integrands (Lepage, J. Comp. Phys. 27, 192)" << endl << endl;

  Dvegas dv5(4, 50, 1, &lepageEq8);
  VEGAS(dv5, 1000, 10);

  cout << endl << "test correlation detector" << endl;
  
  Dvegas dv5corrtest1(4, 50, 1, &lepageEq8);
  dv5corrtest1.activateCorrelationsDetector();
  VEGAS(dv5corrtest1, 250000, 1);

  cout << endl << "test correlation detector" << endl;
  
  Dvegas dv5corrtest2(4, 50, 1, &lepageEq8);
  dv5corrtest2.activateCorrelationsDetector();
  VEGAS(dv5corrtest2, 2500000, 1);

  cout << endl;
  
  Dvegas dv6(4, 1, &crudeLepageEq8);
  VEGAS(dv6, 10000, 2);

  cout << endl;
  
  Dvegas dv7(4, 50, 1, &lepageEq9);
  VEGAS(dv7, 20000, 15);

  cout << endl << "test correlation detector" << endl;
  
  Dvegas dv7corrtest1(4, 50, 1, &lepageEq9);
  dv7corrtest1.activateCorrelationsDetector();
  VEGAS(dv7corrtest1, 250000, 1);

  cout << endl << "test correlation detector" << endl;
  
  Dvegas dv7corrtest2(4, 50, 1, &lepageEq9);
  dv7corrtest2.activateCorrelationsDetector();
  VEGAS(dv7corrtest2, 2500000, 1);

  cout << endl << "+++++ optimizing sums of integrals" << endl << endl;

  Dvegas dv8(1, 50, vector<int>(1, 100), 1, &adaptationForSumsAndAverages);
  VEGAS(dv8, 10000, 4);

  cout << endl;
  
  Dvegas dv8s(1, 50, vector<int>(1, 100), 1, &adaptationForSumsAndAverages, STRATIFIED);
  VEGAS(dv8s, 10000, 4);
  
  cout << endl;
  
  Dvegas dv9(1, 50, 0, vector<int>(), 1, 1, &crudeForSumsAndAverages);
  VEGAS(dv9, 10000, 4);     // sum not optimized, but cont. dim. are optimized

  cout << endl << "+++++ adaptation that takes correlations into account" << endl << endl;

  Dvegas dv10(3, 1, &nastyCrudeMC);
  VEGAS(dv10, 50000, 10);

  cout << endl;

  // classic VEGAS
  Dvegas dv11(3, 30, 0, vector<int>(), 0, 1, &nastyWithAdaptation);
  VEGAS(dv11, 50000, 10);     // note: no improvement over crude MC

  cout << endl;

  // take into account correlations of 2 dimensions
  Dvegas dv13(3, 30, 2, vector<int>(), 0, 1, &nastyWithAdaptation);
  VEGAS(dv13, 50000, 3);     // x-y dependence fully sampled

  cout << endl;
  
  // take into account correlations of 3 dimensions
  Dvegas dv14(3, 30, 3, vector<int>(), 0, 1, &nastyWithAdaptation);
  VEGAS(dv14, 50000, 3);     // x-y-z dependence fully sampled

  /*
  cout << endl << "double Gaussian" << endl;

  Dvegas dv14a(4, 30, 4, vector<int>(), 0, 1, &lepageEq9);
  //dv14a.disableOptimize(0);
  Dvegas::CellAccumulators ca14a = dv14a.collectData(1000000);
  for (int i = 0; i < 4; ++i) {
    dv14a.adapt(ca14a);
    ca14a = dv14a.collectData(20000);     // all correlations importance sampled
  }
  */

  cout << endl << "two sets of correlated dimensions" << endl;
  
  Dvegas dv15(6, 30, vector<int>(1, 3), vector<int>(), 0, 1, &twoSetsOfCorrelatedDimensions);
  //dv15.setCurvature(1.0);
  cout << "curvature = " << dv15.getCurvature() << endl;
  //dv15.setRootAmpl(0.5);
  cout << "rootAmpl  = " << dv15.getRootAmpl() << endl;
  dv15.activateCorrelationsDetector();
  Dvegas::CellAccumulators ca15 = dv15.collectData(10000000);
  for (int i = 0; i < 4; ++i) {
    dv15.adapt(ca15);
    ca15 = dv15.collectData(200000);     // x1-x2-x3 correlation fully sampled
  }

  cout << endl;

  vector<int> corrDim(2);
  corrDim[0] = 3;
  corrDim[1] = 6;
  Dvegas dv16(6, 30, corrDim, vector<int>(), 0, 1, &twoSetsOfCorrelatedDimensions);
  Dvegas::CellAccumulators ca16 = dv16.collectData(10000000);
  for (int i = 0; i < 4; ++i) {
    dv16.adapt(ca16);
    ca16 = dv16.collectData(200000);     // x1-x2-x3 and x4-x5-x6 correlations fully sampled
  }

  cout << endl << "+++++ save and restore state of Dvegas object" << endl << endl;

  dv3.saveStateToFile("dv3.state");
  VEGAS(dv3, 10000, 1, 2);
  dv3.restoreWeightsFromFile(&adaptiveMC, "dv3.state");
  VEGAS(dv3, 10000, 1, 2);     // consistent results for this iteration
  dv3.restoreStateFromFile(&adaptiveMC, "dv3.state");
  VEGAS(dv3, 10000, 1, 2);     // consistent results for this iteration and cumulative
  
  cout << endl;
  
  dv8.saveStateToFile("dv8.state");
  VEGAS(dv8, 10000, 1, 2);
  dv8.restoreStateFromFile(&adaptationForSumsAndAverages, "dv8.state");
  VEGAS(dv8, 10000, 1, 2);

  cout << endl;
  
  dv13.saveStateToFile("dv13.state");
  VEGAS(dv13, 50000, 1, 2);
  dv13.restoreStateFromFile(&nastyWithAdaptation, "dv13.state");
  VEGAS(dv13, 50000, 1, 2);

  return 0;
}

